#ifndef SysLog_H
#define SysLog_H

/*
 * Copyright (c) 2015 Norman J. Goldstein and 
 *                    Clear System Solutions RD Inc.
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 */

/*
  The SysLog_mgr class is a resource manager, in the sense that
  -- The constructor opens a connection to the system log
  -- The destructor closes that connection

  The SysLog global function allows logging to either syslog, cout,
  cerr or clog.  This is run-time switchable via static method

      SysLog_mgr::setOutput( int )

  Also, the SysLog global function, defined as a variadic template,
  has the signature

       bool SysLog( int level, T1 t1, T2 t2, ... )

  The T parameters are treated like

       cout << t1 << t2 << t3 ...

  The direct printf interface of syslog may also be used, but, then,
  redirection to the standard streams is not supported.
 */

#include <syslog.h>
#include <sstream>   // insertion convert to string
#include <ostream>

// Only one instance allowed per process
class SysLog_mgr
{
public:
  SysLog_mgr( const char *ident, int option, int facility)
  {
    if( nullptr == ident )
    {
      openlog( nullptr, option, facility );
    }
    else
    {
      name_ = ident;
      openlog( name_.c_str(), option, facility );
    }
  }

  ~SysLog_mgr() { closelog(); }

  // val can be:
  // 1 : cout
  // 2 : cerr
  // 3 : clog
  // otherwise : syslog
  static void setOutput( int val );

private:
  // man 3 syslog  suggests that ident may not be copied
  // in by the call to openlog().  So, store it here.
  static std::string name_;

  // Use syslog if this is NULL
  static std::ostream* outs_;

  // Levels correspond to  man 3 syslog.
  static const char* levelNames_[];

  static bool strlog( std::stringstream& sls )
  {
    return true;
  }// strlog

  template< typename T, typename... Args > 
  static bool strlog( std::stringstream& sls, T& t, Args... args )
  {
    if( !( sls << t ) )
    {
      return false;
    }

    return strlog( sls, args... );
  }// strlog

  template< typename... Args >
  friend
  bool SysLog( int level, Args... args );

  SysLog_mgr( const SysLog_mgr& other ) = delete;
  SysLog_mgr( SysLog_mgr&& other ) = delete;
  SysLog_mgr& operator=( const SysLog_mgr& other ) = delete;
  SysLog_mgr& operator=( SysLog_mgr&& other ) = delete;  
};

// Levels are 0-7, as in  man 3 syslog.
template< typename... Args >
bool SysLog( int level, Args... args )
{

  if( !( setlogmask(0) & LOG_MASK(level) ) ) return true;
  
  std::stringstream sls;

  if( !SysLog_mgr::strlog( sls, args... ) )
  {
    return false;
  }

  // Check if using syslog
  if( nullptr == SysLog_mgr::outs_ )
  {
    syslog( level, "%s", sls.str().c_str() );
    return true;
  }

  // Using a stream
  *( SysLog_mgr::outs_ ) << SysLog_mgr::levelNames_[ level ]
			 << ": "
			 << SysLog_mgr::name_
			 << ": "
			 << sls.str()
			 << std::endl;

  return SysLog_mgr::outs_->good();
}// SysLog

#endif
