#ifndef resource_mgr_H
#define resource_mgr_H

/*
 * Copyright (c) 2015 Norman J. Goldstein and 
 *                    Clear System Solutions RD Inc.
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 */

/*
  The structs in this file are resource managers.  When the object
  goes out of scope, the resource is freed.  If the struct has the
  method get(), possibly const, it should return the resource that
  is being managed.  The return value can be a value, a refererence or
  a pointer, whichever is appropriate for the particualr resource.
 */


//============ Integer file descriptors ===============
#include <unistd.h>  // close

// Example:
// IntFD fd = open( "filename", O_WRONLY );
struct IntFD_mgr {
  int fd_;
  IntFD_mgr( int fd = -1 ) : fd_( fd ) {}

  int get( void ) const { return fd_; }

  ~IntFD_mgr(){ close( get() ); }

private:
  IntFD_mgr( const IntFD_mgr& other ) = delete;
  IntFD_mgr( IntFD_mgr&& other ) = delete;
  IntFD_mgr& operator=( const IntFD_mgr& other ) = delete;
  IntFD_mgr& operator=( IntFD_mgr&& other ) = delete;  
};


//============ X resources ===============
#include <X11/Xlib.h>

struct Display_mgr
{
  Display* display_;

  Display_mgr( Display* disp = nullptr ) :
    display_( disp )
  {}

  Display*  get( void ) const { return display_; };
  
  ~Display_mgr() { XCloseDisplay( get() ); }

private:
  Display_mgr( const Display_mgr& other ) = delete;
  Display_mgr( Display_mgr&& other ) = delete;
  Display_mgr& operator=( const Display_mgr& other ) = delete;
  Display_mgr& operator=( Display_mgr&& other ) = delete;  
};
	      
struct FontWithDisplay_mgr
{
  XFontStruct* fontStruct_;
  Display* display_;

  FontWithDisplay_mgr( XFontStruct* fs = nullptr,
		       Display* disp = nullptr ) :
    fontStruct_( fs ),
    display_( disp )
  {}

  XFontStruct*  get( void ) const { return fontStruct_; };
  
  ~FontWithDisplay_mgr() { XFreeFont( display_, get() ); }
  
private:
  FontWithDisplay_mgr( const FontWithDisplay_mgr& other ) = delete;
  FontWithDisplay_mgr( FontWithDisplay_mgr&& other ) = delete;
  FontWithDisplay_mgr& operator=( const FontWithDisplay_mgr& other ) = delete;
  FontWithDisplay_mgr& operator=( FontWithDisplay_mgr&& other ) = delete;  
};


//============ XDG management  ===============
#include <basedir.h>

// Example:
// XDGHandle xdgh;
struct XDGHandle_mgr {
  xdgHandle handle_;
  
  XDGHandle_mgr( void ) { xdgInitHandle( get() ); }

  xdgHandle* get( void ) { return &handle_; }

  ~XDGHandle_mgr(){ xdgWipeHandle( get() ); }
  
private:
  XDGHandle_mgr( const XDGHandle_mgr& other ) = delete;
  XDGHandle_mgr( XDGHandle_mgr&& other ) = delete;
  XDGHandle_mgr& operator=( const XDGHandle_mgr& other ) = delete;
  XDGHandle_mgr& operator=( XDGHandle_mgr&& other ) = delete;  
};

#endif
