/*
 * Copyright (c) 2015 Norman J. Goldstein and 
 *                    Clear System Solutions RD Inc.
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 */

/*
  This program may be used as a daemon to suspend and keep the XDG
  screen saver suspended, in order to allow other screen savers to
  work.  The program may also be used to resume (i.e. cancel the
  suspension of) the XDG screen saver.

  During suspension, the daemon process maintains ownership of an X
  window.  The ID of this window is passed to xdg-screensaver to
  effect the suspension of the XDG screen saver.  When the XDG screen
  saver is resumed, the daemon exits, and no resources are being
  consumed (except for a pipe file that remains on disk, ready for the
  next invocation of the daemon).
 */

#include <X11/Xlib.h>
#include <basedir.h>    // XDG

#include <string.h>     // strerror
#include <sys/types.h>  // fifo pipes
#include <sys/stat.h>
#include <fcntl.h>      // O_WRONLY

#include <unistd.h>     // getopt

#include <sstream> 
#include <string>
#include <iostream>

#include "SysLog.h"
#include "resource_mgr.h"

using namespace std;

typedef int (*HANDLER)( Display*, XErrorEvent* );

static int handler( Display* disp, XErrorEvent* xev )
{
  SysLog( LOG_ERR, "Error handler called" );

  return 0;
}// handler

// Path components for creating the pipe file
// in the XDG directory structure.
static const char* subdir = "xdgsr";
static const char* fname = "pipe";

static void usage( void )
{
  cout << R"123(xdgsr flags:
-s     // Suspend xdg screensaver
-r     // Resume xdg screensaver
-h     // Help (this message)
[-oN]  // N = 0 (syslog), 1(cout), 2(cerr), 3(clog)
[-lL]  // L = 0 .. 7 (log up to level L)
Exactly one of the three flags -rsh should be specified.
The -o flag is optional.  The output defaults as
  syslog (suspend)
  cerr (resume)
The -l flag is optional.  
The log level defaults as 5 (log up to NOTICE).
Normal usage is to run   xdgsr -s&   at login, in order
to suspend the xdg screen saver.  The xdgsr daemon process
persists, listening on a pipe.  Should you want to resume
the xdg screensaver, run   xdg -r  .  This closes the daemon.
)123";
}

int main( int argc, char* argv[] )
{
  //========= Parse command line ================
  
  int outval = -1; // Defaults to unspecified
  bool doSuspend = true;
  int logLevel = LOG_NOTICE;
  int flagCount = 0;
  int c = '\0';

  while ( (c = getopt (argc, argv, "rsho:l:")) != -1 )
  {
    switch( c )
    {
      case 'r':
        doSuspend = false;
	++flagCount;
        break;
      case 's':
	doSuspend = true;
	++flagCount;
        break;
      case 'h':
	usage();
        return 0;
      case 'o':
	outval = atoi( optarg );
	break;
      case 'l':
	logLevel = atoi( optarg );
	break;
      default:
	usage();
	return -1;
    }
  }

  if( 1 != flagCount )
  {
    cerr << "Exactly one of -rsh should be specified" << endl;
    usage();
    return -1;
  }

  // If outval has not been specified, give it the default values
  // suspend: 0 (syslog)
  // resume: 2 (cerr)
  if( 0 > outval )
  {
    outval = doSuspend ? 0 : 2;
  }

  SysLog_mgr slog( "xdgsr",
		   LOG_CONS | LOG_NDELAY,
		   LOG_LOCAL1 );

  slog.setOutput( outval );
  setlogmask( LOG_UPTO( logLevel ) );

  //============= Construct pipe file name =============

  // Run-time directory for this utility
  string fullpath;
  {
    XDGHandle_mgr handle;

    if( nullptr == handle.get() )
    {
      SysLog( LOG_ERR, "from xdgInitHandle" );
      return -1;
    }

    // Run-time directory
    const char* rtdir = xdgRuntimeDirectory( handle.get() );

    if( nullptr == rtdir )
    {
      SysLog( LOG_ERR, "from xdgRuntimeDirectory" );
      return -1;
    }
    
    SysLog( LOG_DEBUG, "Run-time dir= ", rtdir );

    fullpath = rtdir;
    ( fullpath += '/' ) += subdir;
  }
  string pipeFile = fullpath + '/' + fname;


  //================= Resume =========================
  if( !doSuspend )
  {
    IntFD_mgr fd( open( pipeFile.c_str(), O_WRONLY | O_NONBLOCK ) );

    if( 0 > fd.get() )
    {
      const int err = errno;

      SysLog( LOG_ERR,
	      strerror(errno),
	      "errno= ", err,
	      "\nError opening pipe file for non-block write: ",
	      pipeFile,
	      "\nFirst run 'xdgsr -s'" );

      if( ENXIO == err )
      {
	SysLog( LOG_ERR, "xdgsr daemon not running?" );
      }
      
      return -1;
    }

    if( 1 != write(fd.get(), "r", 1 ) )
    {
      SysLog( LOG_ERR, "sending resume command" );
      return -1;
    }

    return 0;
  }

  //==================== Suspend ========================

  if( 0 != mkdir( fullpath.c_str(), 0700 ) )
  {
    if( EEXIST == errno )
    {
      SysLog( LOG_INFO, "Pipe directory already exists.  Continuing... " );
    }
    else
    {
      SysLog( LOG_ERR,
	      strerror(errno),
	      "Error creating pipe directory: ",
	      fullpath );
      
      return -1;
    }
  }

  if( 0 != mkfifo( pipeFile.c_str(), 0700 ) )
  {
    if( EEXIST == errno )
    {
      SysLog( LOG_INFO, "Pipe file already exists.  Continuing... " );
    }
    else
    {
     SysLog( LOG_ERR,
	     strerror(errno),
	     "Error creating pipe file: ",
	     pipeFile );
     
      return -1;
    }
  }

  // Check if already suspended
  {
    IntFD_mgr fd ( open( pipeFile.c_str(), O_WRONLY | O_NONBLOCK ) );

    if( 0 <= fd.get() )
    {
      SysLog( LOG_INFO, "Daemon already running" );
      
      return -1;
    }
  }


  HANDLER prevHandler = XSetErrorHandler( handler );

  SysLog( LOG_DEBUG, "Previous handler= 0x",
	  hex,
	  reinterpret_cast<long int>( prevHandler ) );

  // Also cleans up windows and resources attached to the display,
  Display_mgr display( XOpenDisplay( nullptr ) );

  if( nullptr == display.get() )
  {
    SysLog( LOG_ERR,
	    "Error opening display.  Check DISPLAY environment variable" );

    return -1;
  }

  // X window creation & setup from
  // http://www.dreamincode.net/forums/topic/166837-linux-writing-our-first-x-windows-application
  // A bare window is not recognized by XDG.  The window needs to have
  // a degree of resources & visibility -- uncertain.
  Window rootWindow = XDefaultRootWindow( display.get() );
  Visual* visual = DefaultVisual( display.get(), 0 );
  int depth = DefaultDepth( display.get(), 0 );
  
  XSetWindowAttributes frame_attributes;
  frame_attributes.background_pixel = XWhitePixel(display.get(), 0);

  Window window = XCreateWindow( display.get(),
				 rootWindow,
				 0, 0,
				 400, 400,
				 5, depth,
				 InputOutput,
				 visual, CWBackPixel, &frame_attributes );

  XStoreName( display.get(), window, "Hello World Example");

  FontWithDisplay_mgr fontinfo( XLoadQueryFont( display.get(), "10x20"),
				display.get() );
  
  XMapWindow( display.get(), window );

 //-------- Still hex from previous manipulator
  SysLog( LOG_DEBUG, hex, "0x", window );

  SysLog( LOG_DEBUG, dec, window );

  string hexWinID;
  {
    stringstream buf;
    buf << hex << window;
    hexWinID = buf.str();
  }

  // Decimal does not work
  string xdgCmd( "xdg-screensaver suspend 0x" );
  xdgCmd += hexWinID;
  SysLog( LOG_INFO, "xdgCmd= ", xdgCmd );
  
  const int retval = system( xdgCmd.c_str() );

  if( WIFEXITED(retval) )
  {
    const int exitStatus = WEXITSTATUS(retval);
    
    if( 0 != exitStatus )
    {
      SysLog( LOG_ERR, "from xdg-screensaver",
	      "\nChild exit status= ", exitStatus );

      return -1;
    }
  }
  else
  {
    SysLog( LOG_ERR, "system: Child process error" );
    return -1;
  }

  IntFD_mgr fd;

  auto openFD = 
    [ &fd, &pipeFile ] {  
    fd.fd_ = open( pipeFile.c_str(), O_RDONLY );

    if( 0 > fd.get() )
    {
      const int err = errno;
      SysLog( LOG_ERR,
	      strerror(errno),
	      "errno= ", err,
	      "\nError opening pipe file for read: ",
	      pipeFile );

      return false;
    }

    return true;
  };

  if( !openFD() )
  {
    SysLog( LOG_ERR, "Error first open of pipe for reading" );
    return -1;
  }

  char cmd = '\0';
  while( 1 )
  {
    const int retval = read( fd.get(), &cmd, 1 );

    // read can return 0, repeatedly, after a write end
    // of the pipe is closed.  So, reinitialize.
    if( 1 != retval )
    {
      SysLog( LOG_DEBUG, "retval= ", retval );;
      
      if( 0 > retval )
      {
	SysLog( LOG_ERR, strerror(errno),
		"During loop: fd= "
		,fd.get() );
	
      }

      close( fd.get() );
      
      if( !openFD() )
      {
	SysLog( LOG_ERR, "subsequent open of pipe for reading" );
	break;
      }
      continue;
    }
    
    switch( cmd )
    {
      case 'r':
      {
	return 0; // Resume
      }
      default:
      {
	SysLog( LOG_ERR, "Bad command '", cmd, "'" );
	break;
      }
    }
  }

  return -1;
}// main
